@charset "utf-8";
@namespace epub "http://www.idpf.org/2007/ops";

body{
	hyphens: auto;
	-epub-hyphens: auto;
}

p{
	margin: 0;
	text-indent: 1em;
}

hr{
	border: none;
	border-top: 1px solid;
	height: 0;
	margin: 1.5em auto;
	width: 25%;
}

q::before,
q::after{
	content: '';
}

h1,
h2,
h3,
h4,
h5,
h6{
	font-variant: small-caps;
	hyphens: none;
	-epub-hyphens: none;
	page-break-after: avoid;
	page-break-inside: avoid;
	text-align: center;
	margin: 3em 0;
}

blockquote + p,
h2 + p,
h3 + p,
h4 + p,
h5 + p,
h6 + p,
header + p,
hr + p,
ol + p,
ul + p,
table + p,
p:first-child{
	text-indent: 0;
}

cite{
	font-style: normal;
}

abbr{
	border: none;
	white-space: nowrap;
}

blockquote cite{
	display: block;
	font-style: italic;
	text-align: right;
}

blockquote cite i{
	font-style: normal;
}

b,
strong{
	font-variant: small-caps;
	font-weight: normal;
}

header{
	page-break-inside: avoid;
	text-align: center;
}

article > header + *,
section > header + *{
	margin-top: 3em;
}

section[epub|type~="titlepage"] img{
	display: block;
	width: 100%;
	margin: 3em auto auto auto;
}

a[epub|type~="noteref"]{
	font-size: .75em;
	vertical-align: super;
}

section[epub|type~="endnotes"] > ol > li{
	margin: 1em 0;
}

section[epub|type~="colophon"],
section[epub|type~="imprint"]{
	text-align: center;
}

section[epub|type~="colophon"] header,
section[epub|type~="imprint"] header{
	line-height: 0;
	margin-top: 3em;
}

section[epub|type~="colophon"] h2,
section[epub|type~="imprint"] h2{
	display: none;
}

img[epub|type~="z3998:publisher-logo"]{
	max-width: 25%;
}

section[epub|type~="colophon"] p,
section[epub|type~="imprint"] p{
	margin: 1em auto 0 auto;
	text-indent: 0;
}

section[epub|type~="imprint"] p{
	font-size: .75em;
	text-align: justify;
	width: 75%;
}

section[epub|type~="colophon"] p + p::before{
	border-top: 1px solid;
	content: "";
	display: block;
	margin: auto auto 1em auto;
	width: 25%;
}

section[epub|type~="colophon"] a{
	font-variant: small-caps;
}

section[epub|type~="imprint"] a,
section[epub|type~="colophon"] a{
	hyphens: none;
	-epub-hyphens: none;
}

section[epub|type~="copyright-page"] p{
	margin: 1em auto;
	text-indent: 0;
}

section[epub|type~="copyright-page"] blockquote p{
	font-style: italic;
	text-align: left;
	text-indent: 0;
}

section[epub|type~="copyright-page"] blockquote p span{
	display: block;
	padding-left: 1em;
	text-indent: -1em;
}

section[epub|type~="copyright-page"] blockquote br{
	display: none;
}

@supports(display: flex){
	section[epub|type~="imprint"]{
		align-content: center;
		display: flex;
		flex-direction: column;
		justify-content: center;
		min-height: calc(98vh - 3em);
		padding-top: 3em;
	}

	section[epub|type~="imprint"] header{
		margin-bottom: 2em;
		margin-top: 0;
	}

	img[epub|type~="z3998:publisher-logo"]{
		max-width: 25vw;
	}
}
