<?xml version="1.0" encoding="utf-8"?>
<html xmlns="http://www.w3.org/1999/xhtml" xmlns:epub="http://www.idpf.org/2007/ops" epub:prefix="z3998: http://www.daisy.org/z3998/2012/vocab/structure/, se: https://standardebooks.org/vocab/1.0" xml:lang="en-US">
	<head>
		<title>Table of Contents</title>
	</head>
	<body epub:type="frontmatter">
		<nav epub:type="toc">
			<h2 epub:type="title">Table of Contents</h2>
			<ol>
				<li>
					<a href="text/titlepage.xhtml">Titlepage</a>
				</li>
				<li>
					<a href="text/imprint.xhtml">Imprint</a>
				</li>
				<li>
					<a href="text/chapter-1.xhtml"><span epub:type="z3998:roman">I</span>: CHAPTER_TITLE</a>
				</li>
				<li>
					<a href="text/colophon.xhtml">Colophon</a>
				</li>
				<li>
					<a href="text/uncopyright.xhtml">Uncopyright</a>
				</li>
			</ol>
		</nav>
		<nav epub:type="landmarks">
			<h2 epub:type="title">Landmarks</h2>
			<ol>
				<li>
					<a href="text/titlepage.xhtml" epub:type="frontmatter titlepage">Titlepage</a>
				</li>
				<li>
					<a href="text/imprint.xhtml" epub:type="frontmatter imprint">Imprint</a>
				</li>
				<li>
					<a href="text/chapter-1.xhtml" epub:type="bodymatter z3998:fiction">WORK_TITLE</a>
				</li>
				<li>
					<a href="text/colophon.xhtml" epub:type="backmatter colophon">Colophon</a>
				</li>
				<li>
					<a href="text/uncopyright.xhtml" epub:type="backmatter copyright-page">Uncopyright</a>
				</li>
			</ol>
		</nav>
	</body>
</html>
