<?xml version="1.0" encoding="utf-8"?>
<html xmlns="http://www.w3.org/1999/xhtml" xmlns:epub="http://www.idpf.org/2007/ops" epub:prefix="z3998: http://www.daisy.org/z3998/2012/vocab/structure/, se: https://standardebooks.org/vocab/1.0" xml:lang="en-US">
	<head>
		<title>Colophon</title>
		<link href="../css/core.css" rel="stylesheet" type="text/css"/>
		<link href="../css/local.css" rel="stylesheet" type="text/css"/>
	</head>
	<body epub:type="backmatter">
		<section id="colophon" epub:type="colophon">
			<header>
				<h2 epub:type="title">Colophon</h2>
				<img alt="The Standard Ebooks logo" src="../images/logo.svg" epub:type="z3998:publisher-logo"/>
			</header>
			<p><i epub:type="se:name.publication.book">TITLE</i><br/>
			was published in YEAR by<br/>
			<a href="AUTHOR_WIKI_URL">AUTHOR</a>.</p>
			<p>This ebook was produced for the<br/>
			<a href="https://standardebooks.org">Standard Ebooks project</a><br/>
			by<br/>
			<a href="PRODUCER_URL">PRODUCER</a>,<br/>
			and is based on a transcription produced in PG_YEAR by<br/>
			<b class="name">TRANSCRIBER_1</b>, <b class="name">TRANSCRIBER_2</b>, and <a href="https://www.pgdp.net">The Online Distributed Proofreading Team</a><br/>
			for<br/>
			<a href="PG_URL">Project Gutenberg</a><br/>
			and on digital scans available at the<br/>
			<a href="IA_URL">Internet Archive</a>.</p>
			<p>The cover page is adapted from<br/>
			<i epub:type="se:name.visual-art.painting">PAINTING</i>,<br/>
			a painting completed in YEAR by<br/>
			<a href="ARTIST_WIKI_URL">ARTIST</a>.<br/>
			The cover and title pages feature the<br/>
			<b epub:type="se:name.visual-art.typeface">League Spartan</b> and <b epub:type="se:name.visual-art.typeface">Sorts Mill Goudy</b><br/>
			typefaces created in 2014 and 2009 by<br/>
			<a href="https://www.theleagueofmoveabletype.com">The League of Moveable Type</a>.</p>
			<p>The first edition of this ebook was released on<br/>
			<b>January 1, 1900, 12:00 <abbr class="time eoc">a.m.</abbr></b><br/>
			You can check for updates to this ebook, view its revision history, or download it for different ereading systems at<br/>
			<a href="https://standardebooks.org/ebooks/SE_IDENTIFIER">standardebooks.org/ebooks/SE_IDENTIFIER</a>.</p>
			<p>The volunteer-driven Standard Ebooks project relies on readers like you to submit typos, corrections, and other improvements. Anyone can contribute at <a href="https://standardebooks.org">standardebooks.org</a>.</p>
		</section>
	</body>
</html>
