<?xml version="1.0" encoding="utf-8"?>
<html xmlns="http://www.w3.org/1999/xhtml" xmlns:epub="http://www.idpf.org/2007/ops" epub:prefix="z3998: http://www.daisy.org/z3998/2012/vocab/structure/, se: https://standardebooks.org/vocab/1.0" xml:lang="en-US">
	<head>
		<title>Uncopyright</title>
		<link href="../css/core.css" rel="stylesheet" type="text/css"/>
		<link href="../css/local.css" rel="stylesheet" type="text/css"/>
	</head>
	<body epub:type="backmatter">
		<section id="uncopyright" epub:type="copyright-page">
			<h2 epub:type="title">Uncopyright</h2>
			<blockquote epub:type="z3998:verse">
				<p>
					<span>May you do good and not evil.</span>
					<br/>
					<span>May you find forgiveness for yourself and forgive others.</span>
					<br/>
					<span>May you share freely, never taking more than you give.</span>
				</p>
			</blockquote>
			<p>Copyright pages exist to tell you <em>can’t</em> do something. Unlike them, this Uncopyright page exists to tell you, among other things, that the writing and artwork in this ebook are believed to be in the <abbr>U.S.</abbr> public domain. The <abbr>U.S.</abbr> public domain represents our collective cultural heritage, and items in it are free for anyone in the <abbr>U.S.</abbr> to do almost anything at all with, without having to get permission. Public domain items are free of copyright restrictions.</p>
			<p>Copyright laws are different around the world. If you’re not located in the <abbr>U.S.</abbr>, check with your local laws before using this ebook.</p>
			<p>Non-authorship activities performed on public domain items⁠—so-called “sweat of the brow” work⁠—don’t create a new copyright. That means nobody can claim a new copyright on a public domain item for work like digitization, markup, or typography. Regardless, to dispel any possible doubt on the copyright status of this ebook, Standard Ebooks and its contributors release this ebook under the terms in the <a href="https://creativecommons.org/publicdomain/zero/1.0/">CC0 1.0 Universal Public Domain Dedication</a>, thus dedicating to the worldwide public domain all of the work they’ve done on this ebook, including but not limited to metadata, the titlepage, imprint, colophon, this Uncopyright, and any changes or enhancements to, or markup on, the original text and artwork. This dedication doesn’t change the copyright status of the underlying works, which, though believed to already be in the <abbr>U.S.</abbr> public domain, may not yet be in the public domain of other countries. We make this dedication in the interest of enriching our global cultural heritage, to promote free and libre culture around the world, and to give back to the unrestricted culture that has given all of us so much.</p>
			<p>
				<strong>Standard Ebooks and its contributors offer this ebook as-is and make no representations or warranties of any kind concerning this ebook, express, implied, statutory or otherwise, including without limitation warranties of title, merchantability, fitness for a particular purpose, non infringement, or the absence of latent or other defects, accuracy, or the presence or absence of errors, whether or not discoverable, all to the greatest extent permissible under applicable law.</strong>
			</p>
		</section>
	</body>
</html>
